package com.billyang.entrylib.Database;

import com.billyang.entrylib.EntryLib;
import com.billyang.entrylib.MiraiCodeParser.AudioParser;
import com.billyang.entrylib.MiraiCodeParser.ImageParser;

import java.io.File;
import java.sql.*;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * FileCleaner 类
 * 多线程，继承自 Runnable 接口
 * 清理没有用过的图片与语音
 */
class FileCleaner implements Runnable {

    private Thread t;

    EntryLib entrylib;
    private List<File> imageList, audioList;

    /**
     * 构造函数
     * @param entrylib 传递主类
     * @param imageList 提供图片已存在列表
     * @param audioList 提供语音已存在列表
     */
    FileCleaner(EntryLib entrylib, List<File> imageList, List<File> audioList) {
        this.entrylib = entrylib;
        this.imageList = imageList;
        this.audioList = audioList;
    }

    /**
     * 线程运行方法
     */
    public void run() {
        File imageFolder = new File(EntryLib.IMAGES_FOLDER);
        File audioFolder = new File(EntryLib.AUDIOS_FOLDER);
        if(imageFolder.exists()) {
            List<File> tempList = Arrays.asList(imageFolder.listFiles());
            List<File> imageFiles = new ArrayList(tempList);
            tempList = Arrays.asList(audioFolder.listFiles());
            List<File> audioFiles = new ArrayList(tempList);
            HashSet<File> hs1 = new HashSet(imageList), hs2 = new HashSet(imageFiles), hs3 = new HashSet(audioList), hs4 = new HashSet(audioFiles);

            hs2.removeAll(hs1);
            imageFiles.clear();
            imageFiles.addAll(hs2); //得到需要删除的图片文件

            hs4.remove(hs3);
            audioFiles.clear();
            audioFiles.addAll(hs4); //得到需要删除的语音文件

            for(File file: imageFiles) {
                System.gc(); //需要先垃圾回收才能删除
                file.getAbsoluteFile().delete(); //删除图片
                try {
                    Thread.sleep(50);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }

            for(File file: audioFiles) {
                System.gc(); //需要先垃圾回收才能删除
                file.getAbsoluteFile().delete(); //删除语音
                try {
                    Thread.sleep(50);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }

            entrylib.getLogger().info("数据库整理器已清理未使用的图片缓存：总计" + imageFiles.size() + "个图片");
            entrylib.getLogger().info("数据库整理器已清理未使用的图片缓存：总计" + audioFiles.size() + "个语音");
        }
    }

    /**
     * 线程开始方法
     */
    public void start() {
        if(t == null) {
            t = new Thread(this);
            t.start();
        }
    }
}

/**
 * DatabaseAutoArranger 类
 * 自动数据库整理器
 * 继承自 TimerTask 类
 * 定期执行数据库整理任务
 */
public class DatabaseAutoArranger extends TimerTask {

    EntryLib entrylib;
    Connection c;
    Statement stmt;

    /**
     * 构造函数
     * @param entrylib 传递主类
     */
    public DatabaseAutoArranger(EntryLib entrylib) {
        this.entrylib = entrylib;
    }

    /**
     * 查询数据库中某个表是否存在
     * @param stmt 数据库 Statement 对象
     * @param table 表名
     * @return 表的存在性
     */
    public boolean exists(Statement stmt, String table) {
        try {
            stmt.executeQuery("SELECT * FROM " + table + ";");
            return true;
        } catch (SQLException e) {
            return false;
        }
    }

    /**
     * 连接数据库
     * @param database 数据库名
     * @return 连接成功与否
     */
    boolean connect(String database) {
        try {
            Class.forName("org.sqlite.JDBC");
            c = DriverManager.getConnection("jdbc:sqlite:" + EntryLib.DATABASES_FOLDER + database);

            stmt = c.createStatement();
            if(!exists(stmt, "__MAIN_TABLE")) return false;
        } catch( Exception e ) {
            e.printStackTrace();
            return false;
        }

        return true;
    }

    /**
     * 关闭数据库
     */
    void close() {
        try {
            stmt.close();
            c.close();
        } catch( Exception e ) {
            e.printStackTrace();
        }
    }

    /**
     * 整理数据库
     * 共两项工作：
     * - 删除游离表（即主表不存在记录但存在于数据库中的词条表）
     * - 将词条 id 整理成连续正整数
     * @param fileName 数据库名
     * @return 整理状态
     */
    boolean rearrange(String fileName) {
        List<Integer> list = new ArrayList<>(), deleteIdList = new ArrayList<>(); //获取词条id列表
        List<String> titleList = new ArrayList<>(); //获取词条标题列表

        try {
            String sql = "SELECT * FROM __MAIN_TABLE WHERE ALIAS IS NULL;"; //无别名
            ResultSet rs = stmt.executeQuery(sql);
            while(rs.next()) {
                int id = rs.getInt("ID");
                String title = rs.getString("TITLE");
                list.add(id);
                titleList.add(title);
            }
            rs.close();
        } catch( Exception e ) {
            e.printStackTrace();
        }

        try {
            String sql = "SELECT * FROM __MAIN_TABLE WHERE ALIAS IS NOT NULL;"; //别名项
            ResultSet rs = stmt.executeQuery(sql);
            while(rs.next()) {
                int id = rs.getInt("ID");
                String alias = rs.getString("ALIAS");
                if(titleList.contains(alias)) list.add(id); //目标存在，加入list进行整理
                else deleteIdList.add(id); //目标不存在，删除本项
            }
            rs.close();
        } catch( Exception e ) {
            e.printStackTrace();
        }

        for(int id: deleteIdList) {
            String sql = "DELETE FROM __MAIN_TABLE WHERE ID = " + id + ";";

            boolean success = false;
            for(int i = 1; i <= 5; i ++) {
                try {
                    stmt.executeUpdate(sql); //删除失效别名
                    success = true;
                    break;
                } catch( Exception e ) {
                    entrylib.getLogger().error("无法删除失效别名" + id + "，五秒后重试！（" + i + "/5）");
                    e.printStackTrace();

                    try {
                        TimeUnit.SECONDS.sleep(5);
                    } catch (InterruptedException interruptedException) {
                        interruptedException.printStackTrace();
                    }
                }
            }

            if(!success) {
                entrylib.getLogger().error("无法完成" + fileName + "数据库整理工作，即将退出！");
                return false;
            }
        }

        Collections.sort(list);

        try {
            String sql = "SELECT name FROM sqlite_master where type='table' order by name;";
            ResultSet rs = stmt.executeQuery(sql);
            List<Integer> existTables = new ArrayList<>(); //存在的table列表

            while(rs.next()) {
                String name = rs.getString("name");
                if(!name.equals("__MAIN_TABLE") && !name.equals("__VERSION"))existTables.add(Integer.parseInt(name.replace("TABLE_", "")));
            }

            HashSet<Integer> hs1 = new HashSet(list), hs2 = new HashSet(existTables);

            hs2.removeAll(hs1);
            existTables.clear();
            existTables.addAll(hs2); //获取游离表

            for(int tableId: existTables) {
                String table = "TABLE_" + tableId;

                boolean success = false;

                for(int i = 1; i <= 5; i ++) { //删除游离表，尝试五次
                    try {
                        sql = "DROP TABLE " + table + ";";
                        stmt.executeUpdate(sql);
                        success = true;
                        break;
                    } catch (SQLException e) {
                        entrylib.getLogger().error("无法删除游离表表" + table + "，五秒后重试！（" + i + "/5）");
                        e.printStackTrace();

                        try {
                            TimeUnit.SECONDS.sleep(5);
                        } catch (InterruptedException interruptedException) {
                            interruptedException.printStackTrace();
                        }
                    }
                }

                if(!success) {
                    entrylib.getLogger().error("无法完成" + fileName + "数据库整理工作，即将退出！");
                    return false;
                }
            }
        } catch (SQLException e) {
            e.printStackTrace();
        }

        int mex = 1;

        for(int id: list) {
            if(id == mex) { //与最小编号相同，无需整理
                mex ++;
                continue;
            }

            String newTable = "TABLE_" + mex, oldTable = "TABLE_" + id;
            boolean success = false;

            for(int i = 1; i <= 5; i ++) { //复制原表内容至新表，尝试五次
                try {
                    String sql = "CREATE TABLE " + newTable + " AS SELECT * FROM " + oldTable + ";";
                    stmt.executeUpdate(sql);
                    success = true;
                    break;
                } catch (SQLException e) {
                    entrylib.getLogger().error("无法创建新表" + newTable + "，五秒后重试！（" + i + "/5）");
                    e.printStackTrace();

                    try {
                        TimeUnit.SECONDS.sleep(5);
                    } catch (InterruptedException interruptedException) {
                        interruptedException.printStackTrace();
                    }
                }
            }

            if(!success) {
                entrylib.getLogger().error("无法完成" + fileName + "数据库整理工作，即将退出！");
                return false;
            }
            success = false;

            for(int i = 1; i <= 5; i ++) { //修改主表索引，尝试五次
                try {
                    String sql = "UPDATE __MAIN_TABLE SET ID = " + mex + " WHERE ID = " + id + ";";
                    stmt.executeUpdate(sql);
                    success = true;
                    break;
                } catch (SQLException e) {
                    entrylib.getLogger().error("无法创建修改主表关于id" + id + "的索引，五秒后重试！（" + i + "/5）");
                    e.printStackTrace();

                    try {
                        TimeUnit.SECONDS.sleep(5);
                    } catch (InterruptedException interruptedException) {
                        interruptedException.printStackTrace();
                    }
                }
            }

            if(!success) {
                entrylib.getLogger().error("无法完成" + fileName + "数据库整理工作，即将退出！");
                return false;
            }
            success = false;

            for(int i = 1; i <= 5; i ++) { //删除原表，尝试五次
                try {
                    String sql = "DROP TABLE " + oldTable + ";";
                    stmt.executeUpdate(sql);
                    success = true;
                    break;
                } catch (SQLException e) {
                    entrylib.getLogger().error("无法删除原表" + oldTable + "，五秒后重试！（" + i + "/5）");
                    e.printStackTrace();

                    try {
                        TimeUnit.SECONDS.sleep(5);
                    } catch (InterruptedException interruptedException) {
                        interruptedException.printStackTrace();
                    }
                }
            }

            if(!success) {
                entrylib.getLogger().error("无法完成" + fileName + "数据库整理工作，即将退出！");
                return false;
            }

        }

        return true;
    }

    /**
     * 将字符串中的图片 Mirai 码提取出来
     * @param text 字符串
     * @return 图片文件列表
     */
    public List<File> Text2ImageFile(String text) {
        List<File> list = new ArrayList<>();

        Pattern pt = Pattern.compile(ImageParser.regex);
        Matcher mt = pt.matcher(text);

        int start, end = 0;

        while(mt.find()) {
            start = mt.start();
            end = mt.end();

            String imageId = ImageParser.MiraiCode2Id(text.substring(start, end));
            File file = new File(EntryLib.IMAGES_FOLDER, imageId);

            if(file.exists())list.add(file);
        }

        return list;
    }

    /**
     * 将字符串中的语音 Mirai 码提取出来
     * @param text 字符串
     * @return 语音文件列表
     */
    public List<File> Text2AudioFile(String text) {
        List<File> list = new ArrayList<>();

        Pattern pt = Pattern.compile(AudioParser.regex);
        Matcher mt = pt.matcher(text);

        int start, end = 0;

        while(mt.find()) {
            start = mt.start();
            end = mt.end();

            String audioName = AudioParser.MiraiCode2Name(text.substring(start, end));
            File file = new File(EntryLib.AUDIOS_FOLDER, audioName);

            if(file.exists())list.add(file);
        }

        return list;
    }

    List<File> imageList, audioList;

    /**
     * 将数据库中所有包含的图片与语音提取到列表 fileList 中
     */
    void listFile() {
        List<Integer> idList = new ArrayList<>(); //词条表编号列表

        try {
            String sql = "SELECT * FROM __MAIN_TABLE;";
            ResultSet rs = stmt.executeQuery(sql);
            while(rs.next()) {
                int id = rs.getInt("ID");
                String title = rs.getString("TITLE");

                idList.add(id);
                imageList.addAll(Text2ImageFile(title)); //添加标题
                audioList.addAll(Text2AudioFile(title));
            }
            rs.close();
        } catch( Exception e ) {
            e.printStackTrace();
        }

        for(int id: idList) {
            try {
                String sql = "SELECT * FROM TABLE_" + id + ";";
                ResultSet rs = stmt.executeQuery(sql);
                while(rs.next()) {
                    String content = rs.getString("CONTENT");

                    imageList.addAll(Text2ImageFile(content)); //添加内容
                    audioList.addAll(Text2AudioFile(content));
                }
                rs.close();
            } catch( Exception e ) {
                e.printStackTrace();
            }
        }
    }

    /**
     * 定时任务
     * 整理词条库包含的所有数据库
     */
    @Override
    public void run() {
        File file = new File(EntryLib.DATABASES_FOLDER);
        if(!file.exists()) return;

        File[] files = file.listFiles();

        if(files == null) return;

        entrylib.getLogger().info("数据库整理器开始执行整理任务");

        imageList = new ArrayList<>();
        audioList = new ArrayList<>();

        for(File dbFile: files) {
            if(dbFile.isDirectory()) continue;

            String fileName = dbFile.getName();
            if(!fileName.endsWith(".db")) continue; //保证文件是数据库文件

            entrylib.getLogger().info("开始整理数据库" + fileName);

            if(!connect(fileName)) {
                entrylib.getLogger().warning("无法整理数据库" + fileName + "：数据库连接失败！");
                continue;
            }

            if(rearrange(fileName))entrylib.getLogger().info("数据库" + fileName + "整理完成");

            listFile();

            close();
        }

        FileCleaner fileCleaner = new FileCleaner(entrylib, imageList, audioList);
        fileCleaner.start();

        entrylib.getLogger().info("数据库整理器已完成所有整理任务");
    }
}
